"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a Revit rooms geometry extraction functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Utilities.Objects.result import Result
from duHast.Revit.Rooms.rooms import get_all_rooms
from duHast.Data.Objects.Collectors.Properties.Geometry import geometry_polygon_2 as dGeometryPoly
from duHast.Revit.Common.Geometry.curve import splice_adjacent_lines

from Autodesk.Revit.DB import (
    CurveLoop,
    SpatialElementBoundaryOptions,
    SpatialElementBoundaryLocation,
    ElementId,
    Line,
    Arc,
)


def get_room_boundary_loops(
    revit_room,
    spatial_boundary_option=SpatialElementBoundaryOptions(),
    boundary_location=SpatialElementBoundaryLocation.Center,
):
    """
    Returns all boundary loops for a rooms. Default value set to the center
    boundary location.

    Note: Revit will return multiple BoundarySegments if the wall, which bounding a room, has another wall joining it on the opposing site of the room.

    :param revit_room: The room.
    :type revit_room: Autodesk.Revit.DB.Architecture.Room
    :return: List of boundary loops defining the room.
    :rtype: List of lists of Autodesk.Revit.DB.BoundarySegment
    """

    all_boundary_loops = []
    # set up spatial boundary option
    spatial_boundary_option.StoreFreeBoundaryFaces = True
    spatial_boundary_option.SpatialElementBoundaryLocation = boundary_location
    # get loops
    loops = revit_room.GetBoundarySegments(spatial_boundary_option)
    all_boundary_loops.append(loops)
    return all_boundary_loops


def get_points_from_room_boundaries(boundary_loops):
    """
    Returns a list of lists of points representing the room boundary loops.

    - List of Lists because a room can be made up of multiple loops (holes in rooms!)
    - First nested list represents the outer boundary of a room
    - All loops are implicitly closed ( last point is not the first point again!)

    :param boundary_loops: List of boundary loops defining the room.
    :type boundary_loops: List of lists of Autodesk.Revit.DB.BoundarySegment
    :return: A data geometry instance containing the points defining the boundary loop.
    :rtype: :class:`.DataPolygon`
    """

    loop_counter = 0
    has_inner_loops = False
    data_geo_polygon = dGeometryPoly.DataGeometryPolygon2()
    for boundary_loop in boundary_loops:
        for room_loop in boundary_loop:
            p = None  # segment start point
            loop_points = []
            for segment in room_loop:
                p = segment.GetCurve().GetEndPoint(0)
                loop_points.append(p)
            if loop_counter == 0:
                data_geo_polygon.outer_loop = loop_points
            else:
                data_geo_polygon.inner_loops.append(loop_points)
                has_inner_loops = True
            loop_counter += 1
    if not has_inner_loops:
        data_geo_polygon.inner_loops = []
    return data_geo_polygon


def get_2d_points_from_revit_room(revit_room):
    """
    Returns a list of dataGeometry object containing points representing the flattened(2D geometry) of a room in the model.
    List should only have one entry.

    :param revit_room: The room.
    :type revit_room: Autodesk.Revit.DB.Architecture.Room
    :return: A list of data geometry instance containing the points defining the boundary loop.
    :rtype: list of  :class:`.DataGeometry`
    """

    all_room_points = []
    boundary_loops = get_room_boundary_loops(revit_room)
    if len(boundary_loops) > 0:
        room_points = get_points_from_room_boundaries(boundary_loops)
        all_room_points.append(room_points)
    return all_room_points


def get_2d_points_from_all_revit_rooms(doc):
    """
    Returns a list of dataGeometry object containing points representing the flattened(2D geometry) of all the rooms in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of data geometry instances containing the points defining the boundary loop per room.
    :rtype: list of  :class:`.DataGeometry`
    """

    all_room_point_groups = []
    rooms = get_all_rooms(doc)
    for room in rooms:
        room_points = get_2d_points_from_revit_room(room)
        if len(room_points) > 0:
            all_room_point_groups.append(room_points)
    return all_room_point_groups


def convert_boundary_segments_to_curve_loops(boundary_loop):
    """
    Takes a boundary loop and checks whether consecutive segments host have the same id ( same wall  or room separation line. )
    If that is the case it will attempt to splice these segments and return them as one curve within the CurveLoop.

    :param boundary_loop: List of boundary segments defining the room.
    :type boundary_loop: List of Autodesk.Revit.DB.BoundarySegment

    :return:
        Result class instance.

        - conversion status (bool) returned in result.status. False if an exception occurred, otherwise True.
        - Result.message contain logs of segment conversion step by step.
        - Result.result will contain the created CurveLoop instance.

        On exception:

        - .status (bool) will be False.
        - .message will contain the exception message.

    :rtype: :class:`.Result`
    
    """
    return_value = Result()

    # set up curve loops for ceiling creation
    curve_loop = CurveLoop()

    # boundary loops are nested and contain the outlines of the room
   
    current_host_id = ElementId.InvalidElementId
    current_curve = None
    
    counter = 0
    # change boundary loop to curve loop
    for boundary_seg in boundary_loop:

        # get the host id
        host_id = boundary_seg.ElementId
        # get the curve from the boundary segment
        curve = boundary_seg.GetCurve()

        # logging
        return_value.append_message("current host id: {} and host id: {}".format(current_host_id, host_id))

        # only add the previous curve ... in case curves need to be combined!!
        # if curve host as changed, append the current curve to the curve loop
        if host_id != current_host_id:
            return_value.append_message("host id changed: {} to {}".format(current_host_id, host_id))
            return_value.append_message("current curve: {} {} ".format(curve.GetEndPoint(0), curve.GetEndPoint(1)))
            # set the current host id
            current_host_id = host_id
            # if there is a current curve, append it to the curve loop
            if current_curve is not None:
                curve_loop.Append(current_curve)
            # set the new curve to be the current curve
            current_curve = curve
        else:

            # check if line or arc...if neither do not combine
            if not isinstance(curve, Line) and not isinstance(curve, Arc):
                return_value.append_message("current curve is not a line or arc: {}".format(curve))
                # if the curve is not a line or arc, do not combine
                curve_loop.Append(curve)
                current_curve = curve

            # check if curve is a line
            elif isinstance(curve, Line):
                return_value.append_message ("current curve: {} {} and curve: {} {}".format(
                    current_curve.GetEndPoint(0),
                    current_curve.GetEndPoint(1),
                    curve.GetEndPoint(0),
                    curve.GetEndPoint(1))
                )
                return_value.append_message("current curve is a line: {}".format(current_curve))
                combined_curve = splice_adjacent_lines(current_curve, curve)

                if combined_curve is None:
                    return_value.append_message("failed to combine")
                    # if no match is found, do not combine
                    current_curve = curve
                    continue
                
                return_value.append_message("combined curve: {} {}".format(combined_curve.GetEndPoint(0),combined_curve.GetEndPoint(1)))
                # dont append the combined curve in case its more then 2 segments 
                current_curve = combined_curve
            # check if curve is an arc
            elif isinstance(curve, Arc):
                return_value.append_message("current curve is an arc: {}".format(curve))
                # dont combine just yet
                current_curve = curve
            # curve is neither a line nor an arc
            else:
                # if the curve is not a line or arc, do not combine
                return_value.append_message("current curve is not a line or arc: {}".format(curve))
                current_curve = curve
                
        counter = counter + 1
    
    # append the last curve to the curve loop
    if current_curve is not None:
        return_value.append_message("appending last curve: {} {}".format(current_curve.GetEndPoint(0), current_curve.GetEndPoint(1)))
        curve_loop.Append(current_curve)


    return_value.append_message("counter : {}".format(counter))
    return_value.result.append(curve_loop)
    return return_value


# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result

from duHast.Revit.Family.Utility.family_swap_instances_by_type_utils import get_swap_directives
from duHast.Revit.Family.family_swap_instances_of_types import swap_family_instances_of_types
from duHast.Revit.Family.LibraryCleanUp.Utility.family_load import load_families_required_for_swapping
from duHast.Revit.Common.file_io import  sync_file

from duHast.pyRevit.Objects.ProgressPyRevit import ProgressPyRevit


def load_families(doc, directive_directory,  library_path, output, forms):
    """
    Load families required for swapping based on directives.

    :param doc: The Revit document to work with.
    :param directive_directory: Directory containing swap directives.
    :param library_path: Path to the family library.
    :param output: Function to output messages.
    :param forms: Forms module for progress bar.

    :return: Result object containing status and messages.
    """

    return_value = Result()
    try:
        # load swap directives
        swap_directives_result = get_swap_directives(directory_path=directive_directory)

        if not swap_directives_result.status:
            return_value.update_sep(False, "Failed to read swap directives: {}".format(swap_directives_result.message))
            output("Failed to read swap directives: {}".format(swap_directives_result.message))
            return return_value
        
        swap_directives = swap_directives_result.result


        # set up a pyrevit progress bar
        with forms.ProgressBar(
            title="Loading families: {value} of {max_value}", cancellable=True
            ) as pb:

            # set up a call back for pyRevit progressbar
            progress_callback = ProgressPyRevit(form=pb)

            # load families required for swapping
            load_result = load_families_required_for_swapping(
                doc,  
                swap_directives, 
                library_path, 
                progress_callback=progress_callback
            )

            if not load_result.status:
                return_value.update_sep(False, "Failed to load families required for swapping: {}".format(load_result.message))
                #output("Failed to load families required for swapping: {}".format(load_result.message))
                #return return_value
            else:
                return_value.append_message("Families successfully loaded for swapping: {}".format(len(load_result.result)))
            
            output("Families loaded for swapping: {} with status: {}".format(len(load_result.result), load_result.status))

            if( len(load_result.result) == 0):
                return_value.update_sep(True, "No families needing swapping in file")
                output("No families needing swapping in file")
           
    except Exception as e:
        return_value.update_sep(False, "An error occurred: {}".format(str(e)))
        output("An error occurred: {}".format(str(e)))

    return return_value


def swap_families(doc, directive_directory, output,forms):
    """
    Swap family instances in the Revit document based on directives.

    :param doc: The Revit document to work with.
    :param directive_directory: Directory containing swap directives.
    :param output: Function to output messages.
    :param forms: Forms module for progress bar.

    :return: Result object containing status and messages.
    """

    return_value = Result()

    try:
        # set up a pyrevit progress bar
        with forms.ProgressBar(
            title="Swapping family instances: {value} of {max_value}", cancellable=True
            ) as pb:

            # set up a call back for pyRevit progressbar
            progress_callback = ProgressPyRevit(form=pb)
        
            # swap away
            swap_result = swap_family_instances_of_types(doc, directive_directory, progress_callback=progress_callback)

            # print logs
            output("Swapped families with status: {}".format(swap_result.status))

            # check what came back
            if not swap_result.status:
                return_value.update_sep(False, "Failed to swap families: {}".format(swap_result.message))
                output("Failed to swap families: {}".format(swap_result.message))
                return return_value
            return_value.append_message(swap_result.message)
    except Exception as e:
        return_value.update_sep(False, "An error occurred: {}".format(str(e)))
        output("An error occurred: {}".format(str(e)))

    return return_value


def in_process_project(doc, library_path, directive_directory, output, forms):
    """
    Process the Revit document to load families and swap instances based on directives.

    :param doc: The Revit document to work with.
    :param library_path: Path to the family library.
    :param directive_directory: Directory containing swap directives.
    :param output: Function to output messages.
    :param forms: Forms module for progress bar.

    :return: Result object containing status and messages.
    """

    return_value = Result()

    # find families which need swapping load them and swap them out
    # save family

    try:

        # load families
        load_result = load_families(doc, directive_directory, library_path, output, forms)
        return_value.update(load_result)

        # get out if the user cancelled
        if load_result.message.endswith("User cancelled!"):
            return return_value

        # swap families
        swap_result = swap_families(doc, directive_directory, output,forms)
        return_value.update(swap_result)

        # get out if the user cancelled
        if swap_result.message.endswith("User cancelled!"):
            return return_value

        # sync
        #output("Syncing project file.")
        # save the project
        #save_result = sync_file(
        #     doc=doc,
        #     compact_central_file=True)
        
        # output("Save result: {}".format(save_result.status))

    except Exception as e:
        return_value.update_sep(False, "An error occurred: {}".format(str(e)))
        output("An error occurred: {}".format(str(e)))
        return return_value
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A base class used to store category overrides.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


Stores common overrides between categories and filters

"""


#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import json

from duHast.Utilities.Objects import base
from duHast.Revit.Common.Objects.Data import (
    pattern_foreground,
    pattern_background,
    line_projection,
)


class OverrideProjection(base.Base):
    data_type = "override_projection"

    def __init__(self, j=None, **kwargs):
        """
        Class constructor.

        """

        super(OverrideProjection, self).__init__(**kwargs)

        # set defaults
        self.pattern_background = pattern_background.PatternBackground()
        self.pattern_foreground = pattern_foreground.PatternForeground()
        self.line_projection = line_projection.LineProjection()

        # check if any data was past in with constructor!
        if j is not None:
            # check type of data that came in:
            if type(j) == str:
                # a string
                j = json.loads(j)
            elif type(j) == dict:
                # no action required
                pass
            else:
                raise ValueError(
                    "Argument supplied must be of type string or type dictionary"
                )

            # load overrides and throw exception if something is missing!
            try:

                self.pattern_background = pattern_background.PatternBackground(
                    j=j[pattern_background.PatternBackground.data_type]
                )
                self.pattern_foreground = pattern_foreground.PatternForeground(
                    j=j[pattern_foreground.PatternForeground.data_type]
                )
                self.line_projection = line_projection.LineProjection(
                    j=j[line_projection.LineProjection.data_type]
                )
            except Exception as e:
                raise ValueError(
                    "Node {} failed to initialise with: {}".format(
                        OverrideProjection.data_type, e
                    )
                )

    def __eq__(self, other):
        """
        Custom compare is equal override.

        :param other: Another instance of OverrideProjection class
        :type other: :class:`.OverrideProjection`
        :return: True if all properties of compared class instances are equal, otherwise False.
        :rtype: Bool
        """

        return isinstance(other, OverrideProjection) and (
            self.pattern_background,
            self.pattern_foreground,
            self.line_projection,
        ) == (
            other.pattern_background,
            other.pattern_foreground,
            other.line_projection,
        )

    # python 2.7 needs custom implementation of not equal
    def __ne__(self, other):
        return not self.__eq__(other=other)

    def __hash__(self):
        """
        Custom hash override

        Required due to custom __eq__ override present in this class
        """

        try:
            return hash(
                (
                    self.pattern_background,
                    self.pattern_foreground,
                    self.line_projection,
                )
            )
        except Exception as e:
            raise ValueError(
                "Exception {} occurred in {} with values: pattern background:{}, pattern fore: {}, line projection: {}".format(
                    e,
                    self.data_type,
                    self.pattern_background,
                    self.pattern_foreground,
                    self.line_projection,
                )
            )



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit families geometry helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Common.Geometry.solids import (
    get_bounding_box_from_family_geometry,
)
from duHast.Revit.Common.Geometry.geometry import merge_bounding_box_xyz

from Autodesk.Revit.DB import (
    Options,
)


def get_solids_based_bounding_box_from_family_instance(doc, family_instance):
    """
    Returns a bounding box of the family instance based on solids in the family instance geometry only.

    :param doc: The revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param family_instance: The family instance to get the bounding box from.
    :type family_instance: Autodesk.Revit.DB.FamilyInstance

    :return: The bounding box of the family instance or None if no solids within the family instance are found.
    :rtype: Autodesk.Revit.DB.BoundingBoxXYZ or None
    """

    # result is a bounding box
    merged_result = None
    # set a default option
    opts = Options()
    # get the host family geometry first
    merged_result = get_bounding_box_from_family_geometry(
        family_instance.get_Geometry(opts)
    )

    # check if any nested shared families are in play
    sub_element_ids = family_instance.GetSubComponentIds()
    if sub_element_ids is not None:
        # get the bounding box of the sub elements
        for sub_element_id in sub_element_ids:
            sub_element = doc.GetElement(sub_element_id)
            sub_element_bb = get_bounding_box_from_family_geometry(
                sub_element.get_Geometry(opts)
            )
            # check if the result is None and if so set it to the first sub element bounding box
            # this can happen if the host family has no solids but the nested family has
            if merged_result == None:
                merged_result = sub_element_bb
                continue
            # merge the bounding boxes
            merged_result = merge_bounding_box_xyz(merged_result, sub_element_bb)
    return merged_result



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family base data processor class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family.Data.Objects.ifamily_processor import IFamilyProcessor
from duHast.Revit.Family.Data.Objects.family_base_data_processor_defaults import (
    DATA_TYPE_PROCESSOR as data_type_family_base_processor,
)
from duHast.Revit.Family.Data.Objects import family_base_data as rFamData
from duHast.Utilities import util_batch_p as uBP


class FamilyBaseProcessor(IFamilyProcessor):

    data_type = data_type_family_base_processor

    def __init__(
        self,
        reference_file_path=None,
        family_out_directory_path=None,
        session_id=None,
        pre_actions=None,
        post_actions=None,
    ):
        """
        Class constructor.
        """

        # store data type  in base class
        super(FamilyBaseProcessor, self).__init__(
            data_type=FamilyBaseProcessor.data_type,
            pre_actions=pre_actions,
            post_actions=post_actions,
        )

        self.reference_file_path = reference_file_path
        self.family_out_directory_path = family_out_directory_path
        if session_id != None:
            self.session_id = uBP.adjust_session_id_for_directory_name(session_id)
        else:
            self.session_id = session_id

        # self.preActions = preActions
        # self.postActions = postActions

    def process(self, doc, root_path, root_category_path):
        """
        Calls processor instance with the document and root path provided and adds processor instance to class property .data

        :param doc: Current family document.
        :type doc: Autodesk.Revit.DB.Document
        :param rootPath: The path of the nested family in a tree: rootFamilyName :: nestedFamilyNameOne :: nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type rootPath: str
        :param rootCategoryPath: The path of the nested family in in terms of category in a tree: rootFamilyCategory :: nestedFamilyOneCategory :: nestedFamilyTwoCategory\
            This includes the actual family category as the last node.
        :type rootCategoryPath: str
        """

        dummy = rFamData.FamilyBaseData(root_path, root_category_path)
        dummy.process(
            doc,
            self.reference_file_path,
            self.family_out_directory_path,
            self.session_id,
        )
        self.data.append(dummy)



# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#



def clean_up_family_name (fam_name):
    """
    Cleans up the family name by removing  any dashes and spaces, replacing them with underscores.

    :param fam_name: The family name to clean up.
    :type fam_name: str
    :return: The cleaned up family name.
    :rtype: str
    """
    
    if fam_name is None:
        return None
    
    # replace any "-" with "_"
    fam_name = fam_name.replace('-', '_')

    # replace any spaces with "_"
    fam_name = fam_name.replace(' ', '_')

    # remove illegal filename characters
    illegal_chars = ['\\', '/', ':', '*', '?', '"', '<', '>', '|']
    for char in illegal_chars:
        fam_name = fam_name.replace(char, '')
    
    # remove ampersand (&) and replace it with "and"
    fam_name = fam_name.replace('&', 'And')

    return fam_name

def build_family_name_from_descriptor(description, code_to_use, output):
    """
    Builds a family name from the given description by cleaning it up.

    Description is in format

    MAJOR CATEGORY: description part, description part, description part

    :param description: The description to build the family name from.
    :type description: str
    :return: The cleaned up family name.
    :rtype: str
    """
    
    # isolate the major category and change to sentence case


    if description is None:
        return None
    
    # split the description into parts

    #output ("...Building family name from description: {}".format(description))

    parts = description[1].split(':')

    #output("...Parts after splitting by colon: {}".format(parts))

    if len(parts) < 2:
        # just a major category in the descriptor
        fam_name = ("{}_".format(description[1].title().strip().replace(' ', '')))
        return "{}{}".format(fam_name, code_to_use)
        #raise ValueError("Description must contain a major category followed by a colon. {}".format(description))
    
    # get the major category and clean it up
    major_category = parts[0].title().strip().replace(' ', '')
    description_parts = []
    try:
        
        # get the description parts and clean it up: break up in parts at each comma, Capitalize case each part, remove all spaces in each part, join parts by underscore
        for part in parts[1].split(','):
            
            # strip leading/trailing spaces from the part
            part = part.strip()

            # break up into words at each space
            words = []
            # check each word in the part
            for word in part.split(" "):
                # check if part is all upper case idf not capitalize it
                if not word.isupper():
                    #print("word '{}' is not all upper case, capitalizing it.".format(word))
                    word = word.title()  # Capitalize the first letter of each word and strip leading/trailing spaces
                else:
                    #print("Word '{}' is all upper case, keeping it as is.".format(word))
                    pass
                words.append(word)


            description_parts.append("".join(words))
    except Exception as e:
        output("...Error processing description parts: {}".format(e))
        return None
           
    # join the major category and description parts with an underscore
    family_name = major_category + '_' + '_'.join(description_parts)

    # remove any duplicate underscores
    family_name = family_name.replace('__', '_')
    # remove illegal characters from the family name
    family_name = clean_up_family_name (family_name)
   

    #output("...Family name after processing description: {}".format(family_name))
    # add the code to the family name
    # check if family end on underscore
    if family_name.endswith('_'):
        # just append the code
        family_name = "{}{}".format(family_name, code_to_use)
    else:
        # add an underscore before the code
        family_name = "{}_{}".format(family_name, code_to_use)
    
    output("...Family name after processing description and code {} added: {}".format(code_to_use, family_name))
    return family_name


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Model phase functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Common.parameter_get_utils import get_built_in_parameter_value, get_parameter_value_as_element_id

from Autodesk.Revit.DB import BuiltInParameter


def get_all_phases(doc):
    """
    Returns a dictionary where key is the id and value is the name of the phase.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A dictionary where key is the id and value is the name of the phase. Dictionary will be empty if no phases exist (family doc?)
    :rtype: dic{key Autodesk.Revit.DB.ElementId: value str}
    """

    return_value = {}
    phases = doc.Phases
    if phases.Size > 0:
        for phase in phases:
            return_value[phase.Id] = phase.Name
    return return_value


def get_all_phases_in_order(doc):
    """
    Returns all phases in the order of oldest to newest.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of tuples ordered oldest to newest. Tuple properties are: 0 is the phase id and 1 is the phase name.
    :rtype: [(ElementId, str)]

    """
    return_value = []
    phases = doc.Phases
    if phases.Size > 0:
        for phase in phases:
            return_value.append((phase.Id, phase.Name))
    return return_value


def get_phase_name_by_id(doc, phase_id):
    """
    Returns the name of a phase by Id

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param phase_id: The phase element id.
    :type phase_id: Autodesk.Revit.DB.ElementId

    :return: Name of the phase, otherwise 'No phase in document' if no phase exists or 'Invalid phase id.' if no phase matching the id was found.
    :rtype: str
    """

    return_value = "No phase in document"
    phases = get_all_phases(doc)
    if len(phases) > 0:
        if phase_id in phases:
            return_value = phases[phase_id]
        else:
            return_value = "Invalid phase id."
    return return_value


def get_name_to_phase_dict(rvt_doc):
    """
    Returns a dictionary where key is the name and value is the phase.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A dictionary where key is the name and value is the phase. Dictionary will be empty if no phases exist (family doc?)
    :rtype: dic{key str: value Autodesk.Revit.DB.Phase}
    """
    phases = rvt_doc.Phases
    phase_dict = {}

    for phase in phases:
        phase_dict[phase.Name] = phase

    return phase_dict


def get_phase_id_created(element):
    """
    Returns the id of the phase an element was created in.

    :param element: A Revit element
    :type element: Autodesk.Revit.DB.Element

    :return: Phase id
    :rtype: Autodesk.Revit.DB.ElementId
    """

    id = get_built_in_parameter_value(
        element,
        BuiltInParameter.PHASE_CREATED,
        get_parameter_value_as_element_id,
    )
    return id


def get_phase_id_demolished(element):
    """
    Returns the id of the phase an element was demolished in.

    :param element: A Revit element
    :type element: Autodesk.Revit.DB.Element

    :return: Phase id
    :rtype: Autodesk.Revit.DB.ElementId
    """
    
    id = get_built_in_parameter_value(
        element,
        BuiltInParameter.PHASE_DEMOLISHED,
        get_parameter_value_as_element_id,
    )
    return id



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to sorting Revit building pads by types. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.BuildingPads.Utility import (
    RevitBuildingPadsFilter as rBuildingPadFilter,
)


def build_building_pad_type_dictionary(collector, dic):
    """
    Returns the dictionary past in with keys and or values added retrieved from collector past in.
    Keys are built in building pad family type names.
    TODO: Use more generic code.
    :param collector: A filtered element collector containing building pad types.
    :type collector: Autodesk.Revit.DB.FilteredElementCollector
    :param dic: A dictionary containing key:  building pad type family name, value: list of ids belonging to that type.
    :type dic: dictionary (key str, value list of Autodesk.Revit.DB.ElementId)
    :return: A dictionary containing key: built in building pad type family name, value: list of ids belonging to that type.
    :rtype: dictionary (key str, value list of Autodesk.Revit.DB.ElementId)
    """

    for c in collector:
        if c.FamilyName in dic:
            if c.Id not in dic[c.FamilyName]:
                dic[c.FamilyName].append(c.Id)
        else:
            dic[c.FamilyName] = [c.Id]
    return dic


def sort_building_pad_types_by_family_name(doc):
    """
    Returns a dictionary of all building pad types in the model where key is the build in wall family name, values are ids of associated wall types.
    TODO: Use more generic code.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A dictionary containing key: built in building pad type family name, value: list of ids belonging to that type.
    :rtype: dictionary (key str, value list of Autodesk.Revit.DB.ElementId)
    """

    # get all building pad Type Elements
    wts = rBuildingPadFilter._get_building_pad_types_by_class(doc)
    # get all pad types including in place pad families
    wts_two = rBuildingPadFilter._get_all_building_pad_types_by_category(doc)
    usedWts = {}
    usedWts = build_building_pad_type_dictionary(wts, usedWts)
    usedWts = build_building_pad_type_dictionary(wts_two, usedWts)
    return usedWts



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family Base data analysis module containing functions to find missing families.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A missing family is a family which is not present in the family base data report as a root family but present as a nested family.

Algorithm description:

- read base data file processing list

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.timer import Timer
from duHast.Utilities.Objects import result as res
from duHast.Revit.Family.Data.family_data_family_processor_utils import process_data
from duHast.Revit.Family.Data.Objects.family_base_data_processor_defaults import (
    NESTING_SEPARATOR,
)


def get_unique_nested_families_from_path_data(path_data):
    """
    Function to retrieve unique nested families from a list of tuples representing longest unique name nesting path and matching longest unique category nesting path.

    :param path_data: list of tuples representing longest unique name nesting path and matching longest unique category nesting path
    :type path_data: list[(family_name_nesting, family_category_nesting)]
    :return: list of tuples containing unique family name and category
    :rtype: list[(family_name, family_category)]
    """
    # path data is a list of tuples list[(family_name_nesting, family_category_nesting)] or None

    # will be a list of tuples 0: family name, 1 family category
    unique_nested_families = []

    # loop over all unique nesting path
    for entry in path_data:
        family_name_nesting = entry[0]
        category_name_nesting = entry[1]

        # split into chunks at separator
        families = family_name_nesting.split(NESTING_SEPARATOR)
        categories = category_name_nesting.split(NESTING_SEPARATOR)

        if len(families) != len(categories):
            raise ValueError(
                "Name path length: {} is different to category path length: {}".format(
                    len(families), len(categories)
                )
            )
        # loop over each entry in path ignoring the first entry (root family)
        for i in range(1, len(families)):
            test_value = (families[i], categories[i])
            if test_value not in unique_nested_families:
                unique_nested_families.append(test_value)

    return unique_nested_families


def get_unique_root_families_from_family_data(family_data):
    """
    Function to retrieve unique root families from a list of family data objects.

    :param family_data: list of family data objects
    :type family_data: list[:class:`.FamilyDataFamily`]
    :return: list of tuples containing family name and category
    :rtype: list[(family_name, family_category)]
    """

    # family data is a list of family_data_family instances

    # will be a list of tuples 0: family name, 1 family category
    unique_root_families = []

    for family in family_data:
        test_data = (family.family_name, family.family_category)
        if test_data not in unique_root_families:
            unique_root_families.append(test_data)

    return unique_root_families


def get_missing_families(root_families, nested_families):
    """
    Function to find missing families from a list of root families and nested families.

    :param root_families: list of tuples representing root family name and category
    :type root_families: list[(family_name, family_category)]
    :param nested_families: list of tuples representing nested family name and category
    :type nested_families: list[(family_name, family_category)]
    :return: list of tuples representing nested family name and category which does not have a matching root family
    :rtype: list[(family_name, family_category)]
    """

    missing_families = []

    for nested in nested_families:
        if nested not in root_families:
            missing_families.append(nested)

    return missing_families


def process_families(family_data, result_list):
    """
    Function to retrieve longest unique nesting path data from families.

    :param family_data: list of family data objects
    :type family_data: list
    :param result_list: list to store longest unique path
    :type result_list: list[(family, (family_name_nesting, family_category_nesting))]
    :return: list of tuples containing family data object at 0 and longest path at 1 (multiple entries per family possible). Longest path itself is a tuple of family name and category.
    """

    # loop over all family data
    for family in family_data:
        # process each family
        family.process()
        longest_path = (
            family.get_longest_unique_nesting_path()
        )  # returns a list of tuples (family root name path, family root category path)
        if longest_path is not None:
            for lp in longest_path:
                result_list.append((family, lp))
    return result_list


def _find_missing_families(families, families_longest_path):
    """
    Returns a list of tuples representing nested family name and category which does not have a matching root family.

    :param families: List of family instances
    :type families: [:class:`.FamilyDataFamily`]
    :param families_longest_path: list of tuples representing longest unique name nesting path and matching longest unique category nesting path
    :type families_longest_path: [(family name, family category)]
    :return: List of tuples representing the name and category of a family missing (from the library and therefore not presented as root family)
    :rtype: [(family name, family category)]
    """

    missing_families = []
    # process longest path retrieved into unique families
    unique_nested_families = get_unique_nested_families_from_path_data(
        families_longest_path
    )

    # process family data and get unique names+ categories
    unique_root_families = get_unique_root_families_from_family_data(families)

    # check which nested families do not exist as root families
    missing_families = get_missing_families(
        root_families=unique_root_families, nested_families=unique_nested_families
    )

    return missing_families


def check_families_missing_from_library(family_base_data_report_file_path):
    """
    Processes a family base data report and identifies any nested families which have not been processed as a root family\
        and therefore do not exist in the library.
    
    :param family_base_data_report_file_path: Fully qualified file path to family base data report file. 
    :type family_base_data_report_file_path: str
    
    :return: 
        Result class instance.

        - result.status. True if missing families where found without an exception occurring.
        - result.message will contain the summary messages of the process including time stamps.
        - result.result [(nested_family_name, nested_family_category)]
        
        On exception:
        
        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    # read families into data family objects
    # check for duplicate family names in nesting tree

    return_value = res.Result()

    # set up a timer
    t_process = Timer()
    t_process.start()

    try:

        return_value.append_message("Checking for missing families...")
        # load and process families
        families_processed = process_data(
            family_base_data_report_file_path=family_base_data_report_file_path,
            do_this=process_families,
        )

        # check if processing was successful, otherwise get out
        if families_processed.status == False:
            raise ValueError(families_processed.message)

        # get results
        families = []  # list of family instances
        families_longest_path = (
            []
        )  # list of tuples representing longest unique name nesting path and matching longest unique category nesting path
        for nested_tuple in families_processed.result:
            # per nested path there might be multiple entries of the same family
            families.append(nested_tuple[0])
            families_longest_path.append(nested_tuple[1])

        return_value.append_message(
            "{} Found: {} unique longest path in families.".format(
                t_process.stop(), len(families_longest_path)
            )
        )

        # start timer again
        t_process.start()

        # get tuples representing missing root family name and Revit category
        missing_families = _find_missing_families(
            families=families, families_longest_path=families_longest_path
        )

        return_value.append_message(
            "Found {} missing families. {}".format(
                len(missing_families), t_process.stop()
            )
        )
        if len(missing_families) > 0:
            return_value.result = missing_families

    except Exception as e:
        return_value.update_sep(
            False,
            "An error occurred while finding missing families: {}".format(e),
        )

    return return_value


# ----------------------------missing families: direct host files -----------------------------------------


def get_direct_root_families(families, missing_families):
    """
    Returns a list of FamilyDataFamily instances which represent the direct parents (host families) of the missing families.

    :param families: List of family instances
    :type families: [:class:`.FamilyDataFamily`]
    :param missing_families: List of tuples representing the name and category of a family missing (from the library and therefore not presented as root family)
    :type missing_families: [(family name, family category)]
    :return: List of family instances which represent the direct parents (host families) of the missing families
    :rtype: [:class:`.FamilyDataFamily`]
    """

    # return value
    direct_host_families = []
    direct_host_families_short = []
    # loop over families and check for match at nesting level 01
    for family in families:
        # families at nesting level 1
        if 1 in family.nesting_by_level:
            for family_at_level_one in family.nesting_by_level[1]:
                # build test value: a tuple made up of the family name and family category
                test_value = (
                    family_at_level_one.family_name,
                    family_at_level_one.family_category,
                )
                if (
                    test_value in missing_families
                    and "{} {}".format(family.family_name, family.family_category)
                    not in direct_host_families_short
                ):
                    # match found...store the direct host family
                    direct_host_families.append(family)
                    direct_host_families_short.append(
                        "{} {}".format(family.family_name, family.family_category)
                    )

    return direct_host_families


def find_missing_families_direct_host_families(family_base_data_report_file_path):
    """
    Returns a list of FamilyDataFamily instances which represent the direct parents (host families) of the missing families.
    Only processed root families which have a missing family as a direct nested family are returned. (any root families which contains a missing family nested further down the nesting tree are not returned)

    :param family_base_data_report_file_path: Fully qualified file path to family base data report file.
    :type family_base_data_report_file_path: str

    :return:
        Result class instance.

        - result.status. True if host families of missing families where found without an exception occurring.
        - result.message will contain the summary messages of the process including time stamps.
        - result.result [:class:`.FamilyDataFamily`]

        On exception:

        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    # read families into data family objects
    # check for duplicate family names in nesting tree

    return_value = res.Result()

    # set up a timer
    t_process = Timer()
    t_process.start()

    try:

        # load and process families
        families_processed_result = process_data(
            family_base_data_report_file_path=family_base_data_report_file_path,
            do_this=process_families,
        )

        # check if processing was successful, otherwise get out
        if families_processed_result.status == False:
            raise ValueError(families_processed_result.message)

        # get results
        families = []
        families_longest_path = []
        for nested_tuple in families_processed_result.result:
            # per nested path there might be multiple entries of the same family
            if nested_tuple[0] not in families:
                families.append(nested_tuple[0])
            families_longest_path.append(nested_tuple[1])

        return_value.append_message(
            "{} Found: {} unique longest path in families.".format(
                t_process.stop(), len(families_longest_path)
            )
        )

        # start timer again
        t_process.start()

        # get tuples representing missing root family name and Revit category
        missing_families = _find_missing_families(
            families=families, families_longest_path=families_longest_path
        )

        # check if there are any missing families
        if len(missing_families) == 0:
            return_value.append_message(
                "No missing families found in data set. {}".format(t_process.stop())
            )
            return return_value

        # get the direct root families of nested families identified as missing
        # logging
        return_value.append_message(
            "Found {} missing families. {}".format(
                len(missing_families), t_process.stop()
            )
        )
        for mf in missing_families:
            return_value.append_message("Missing family: {} {}".format(mf[0], mf[1]))

        # start timer again
        t_process.start()

        # set up a list for direct root families
        direct_root_families = []

        # loop over longest path and find the ones where the second entry in the nesting path is a missing family
        direct_root_families = get_direct_root_families(
            families=families,
            missing_families=missing_families,
        )

        # logging
        return_value.append_message(
            "Found {} direct hosts to missing families. {}".format(
                len(direct_root_families), t_process.stop()
            )
        )
        for drf in direct_root_families:
            return_value.append_message(
                "Direct host family: {} {}".format(drf.family_name, drf.family_category)
            )

        # update result property as required
        if len(direct_root_families) > 0:
            return_value.result = direct_root_families

    except Exception as e:
        return_value.update_sep(
            False,
            "An error occurred while finding missing families direct host families: {}".format(
                e
            ),
        )

    return return_value



from duHast.Revit.Warnings.Objects.warnings_overlap_storage import RevitWarningOverlap
from duHast.Revit.Warnings.warnings import get_warnings_by_guid
from duHast.Revit.Common.Geometry.curve import (
    is_curve_is_within_curve,
    calculate_lengthened_curve_geometry,
    calculate_shortened_curve_geometry,
)
from duHast.Revit.Common.groups import get_model_group_instances_by_type
from duHast.Revit.Common.delete import delete_by_element_ids
from duHast.Revit.Common.transaction import in_transaction
from duHast.Revit.Purge.purge_unused_by_delete import purge_unused_elements
from duHast.Utilities.Objects import result as res
from duHast.Revit.Rooms.Objects.RoomSeparationLinesPurgeModifier import RoomSeparationLinesPurgeModifier

from Autodesk.Revit.DB import (
    BuiltInCategory,
    Category,
    Document,
    ElementId,
    Transaction,
    ModelLine,
)


def get_curves_from_failure_messages(doc, failure_messages, group_id):
    """
    Loops over failure messages and extracts separation lines listed.
    There are always two curves per exception, even if one curve overlaps a number of other curves.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param failure_messages: List of failure messages relating to separation lines overlap
    :type failure_messages: [Autodesk.Revit.DB.FailureMessage]
    :param group_id: group id curves need to belong to for processing (-1 not in a group)
    :type group_id: int
    :return: List of RevitWarningOverlap instances.
    :rtype: [:class:`RevitWarningOverlap`]
    """

    if not isinstance(doc, Document):
        raise TypeError("doc must be an instance of Autodesk.Revit.DB.Document")
    if not isinstance(failure_messages, list):
        raise TypeError("failure_messages must be a list")

    revit_warnings_overlap_storage = []
    for f_message in failure_messages:
        ids = f_message.GetFailingElements()
        # area line overlap warnings are also covering other overlap warnings which have more then 2 elements
        # check that elements involved are ModelLines ( room sep, area sep)
        # TODO: use builtincategory instead! OST_AreaPolylines, OST_RoomPolylines
        if len(ids) == 2:
            lines_by_exception = []
            are_matching_group_id = True
            for id in ids:
                sep_line = doc.GetElement(id)
                cat = sep_line.Category
                if (
                    cat.Name
                    == Category.GetCategory(
                        doc, BuiltInCategory.OST_AreaSchemeLines
                    ).Name
                    or cat.Name
                    == Category.GetCategory(
                        doc, BuiltInCategory.OST_RoomSeparationLines
                    ).Name
                ):
                    curve = sep_line.GeometryCurve
                    group_id_curve = sep_line.GroupId.IntegerValue
                    if group_id_curve != group_id:
                        are_matching_group_id = False

                    design_option_id = -1
                    if sep_line.DesignOption:
                        design_option_id = sep_line.DesignOption.Id

                    created_phase_id = sep_line.CreatedPhaseId
                    demolished_phase_id = sep_line.DemolishedPhaseId

                    dummy = RevitWarningOverlap(
                        id=id,
                        element=sep_line,
                        curve=curve,
                        group_id=group_id_curve,
                        design_option_id=design_option_id,
                        created_phase_id=created_phase_id,
                        demolished_phase_id=demolished_phase_id,
                    )
                    lines_by_exception.append(dummy)

            # only add warnings if both elements belong to the same group (or both are not in a group)
            # if elements are of type model line ( if not lines_by_exception will be empty )
            if are_matching_group_id and len(lines_by_exception) == 2:
                revit_warnings_overlap_storage.append(lines_by_exception)

    return revit_warnings_overlap_storage


def check_curves_overlaps(curves, group_id):
    """
    Check a list of curve sets for overlaps or duplicates.


    :param curves: A list of curve sets, where each set contains two curves.
    :type  curves: [:class: `RevitWarningOverlap`]
    :param group_id: group id curves need to belong to for processing (-1 not in a group)
    :type group_id: int

    Returns:
        list: A list of curves that can be deleted because they are either completely within another curve or identical to another curve.

    Raises:
        ValueError: If a curve set does not have a length of 2.

    Example:
        curves = [[curve1, curve2], [curve3, curve4], ...]
        result = check_curves_overlaps(curves)
        print(result)
    """

    if not isinstance(curves, list):
        raise TypeError("curves must be a list")
    if not isinstance(group_id, int):
        raise TypeError("group_id must be an instance of integer")

    curves_to_delete = []
    # a curve set should always contain 2 curves
    for curve_set in curves:
        if len(curve_set) == 2:
            # currently only works when separation lines are outside a group..
            # if inside a group the lines get excluded from a group but not actually deleted

            curve_to_delete = is_curve_is_within_curve(curve_set[0], curve_set[1])
            # can anything be deleted?
            # need to be in main model, same design option, same phase created and demoed
            if curve_to_delete:
                if (
                    curve_to_delete.group_id == group_id
                    and curve_set[0].design_option_id == curve_set[1].design_option_id
                    and curve_set[0].phase_created_id == curve_set[1].phase_created_id
                    and curve_set[0].phase_demolished_id
                    == curve_set[1].phase_demolished_id
                ):
                    # check if the curve to be deleted is in the main model
                    curves_to_delete.append(curve_to_delete)
        else:
            raise ValueError("Curve set does not have a length of 2")
    return curves_to_delete


def _identify_curve_in_set_to_amend_lengthening(curve_set, group_id):
    """
    Returns two lists:
    First one contains curves to change the geometry of
    Second one contains curves to delete

    :param curve_set: A set of curves that needs to be analyzed and identified for amendment
    :type curve_set: [:class: `RevitWarningOverlap`]
    :return: A tuple containing two lists - curves_to_amend and curves_to_delete
    :param group_id: group id curves need to belong to for processing (-1 not in a group)
    :type group_id: int
    :rtype: tuple
    :raises ValueError: If the length of curve_set is not equal to 2
    """

    if not isinstance(curve_set, list):
        raise TypeError("curve_set must be a list")
    if not isinstance(group_id, int):
        raise TypeError("group_id must be an instance of integer")

    curves_to_delete = []
    curves_to_amend = []

    if len(curve_set) == 2:
        # calculate changes
        curve_to_change, curve_to_delete = calculate_lengthened_curve_geometry(
            curve_set[0], curve_set[1]
        )

        # check whether changes can be executed
        # need to be in main model, same design option, same phase created and demoed
        if (
            curve_to_delete
            and curve_to_change
            and curve_to_delete.group_id == group_id
            and curve_to_change.group_id == group_id
            and curve_set[0].design_option_id == curve_set[1].design_option_id
            and curve_set[0].phase_created_id == curve_set[1].phase_created_id
            and curve_set[0].phase_demolished_id == curve_set[1].phase_demolished_id
        ):
            # check if the curve to be deleted is in the main model, that there us a curve to lengthen and its also in main model
            curves_to_delete.append(curve_to_delete)
            curves_to_amend.append(curve_to_change)
        else:
            return None, None
    else:
        raise ValueError("Curve set does not have a length of 2")

    return curves_to_amend, curves_to_delete


def delete_curves(doc, curves_to_delete, curve_descriptor, progress_callback=None):
    """
    Deletes curves in a Revit model.

    :param doc: The Revit model document.
    :type doc:  (Autodesk.Revit.DB.Document)
    :param curves_to_delete: List of curves to be deleted.
    :type curves_to_delete: [:class: `RevitWarningOverlap`]
    :param curve_descriptor: line type descriptor
    :type curve_descriptor: str

    Returns:
        res.Result: A res.Result object that contains the status of the deletion operation.
    """

    if not isinstance(doc, Document):
        raise TypeError("doc must be an instance of Autodesk.Revit.DB.Document")
    if not isinstance(curves_to_delete, list):
        raise TypeError("curves_to_delete must be a list")

    return_value = res.Result()
    # get unique ids for deletion
    if len(curves_to_delete) > 0:

        # set up element Id getter
        def action(doc):
            ids = []
            ids = [
                c.id for c in curves_to_delete if c and c.id is not None and c.id not in ids
            ]
            return ids

        # set up a modified modifier ( filter out any modified rooms)
        modifier_modified_rooms = RoomSeparationLinesPurgeModifier(doc)
        # purge unused curves only if nothing else is affected
        purge_result = purge_unused_elements(
            doc=doc,
            element_id_getter=action,
            deleted_elements_modifier=modifier_modified_rooms,
            modified_elements_modifier=modifier_modified_rooms,
            progress_callback=progress_callback,
            debug=False,
        )

        return_value.update(purge_result)
    else:
        return_value.update_sep(True, "No curves where required to be deleted.")
    return return_value


def _modify_separation_lines(doc, curves, transaction_manager=in_transaction):
    """
    Modify the geometry of separation lines in a Revit model.

    :param doc: The Revit model document.
    :type doc:  (Autodesk.Revit.DB.Document)
    :param curves: List of curves to be deleted.
    :type curves: [:class: `RevitWarningOverlap`]
    :param  transaction_manager:  An optional transaction manager function used to perform the modifications.
    :type  transaction_manager: (function, optional)

    Returns:
        res.Result: A `res.Result` object that contains the status of the modifications performed for each curve.
    """

    if not isinstance(doc, Document):
        raise TypeError("doc must be an instance of Autodesk.Revit.DB.Document")
    if not isinstance(curves, list):
        raise TypeError("curves must be a list")

    return_value = res.Result()
    for curve in curves:
        if curve:

            def action():
                action_return_value = res.Result()
                try:
                    # make sure to set this to True, otherwise adjoining room separation lines will change as well
                    curve.element.SetGeometryCurve(curve.new_curve, True)
                    action_return_value.update_sep(
                        True,
                        "Successfully change geometry of element: {}".format(curve.id),
                    )
                except Exception as e:
                    action_return_value.update_sep(
                        False,
                        "Failed to change geometry of element {} with exception: {}".format(
                            curve.id, e
                        ),
                    )
                return action_return_value

            transaction = Transaction(
                doc, "Updating separation line geometry: {}".format(curve.id)
            )
            update_curve_geometry = transaction_manager(transaction, action)
            # update_sep_geo = in_transaction(transaction, action)
            return_value.update(update_curve_geometry)
    return return_value


def modify_curves_by_lengthening(doc, guid, transaction_manager, group_id, callback):
    """
    Modifies curves in a Revit model by lengthening them.

    :param doc: The Revit model document.
    :type doc:  (Autodesk.Revit.DB.Document)
    :param guid: The GUID used to filter the failure messages.
    :type guid: (Autodesk.Revit.DB.Guid):
    :param transaction_manager: The transaction manager used to perform the modifications.
    :type transaction_manager: (function)
    :param group_id: group id curves need to belong to for processing (-1 not in a group)
    :type group_id: int
    :param callback: function reporting progress update.
    :type callback: func (value, maximum_value)

    Returns:
        res.Result: A `res.Result` object that contains the status of the modifications performed.
    """

    if not isinstance(doc, Document):
        raise TypeError("doc must be an instance of Autodesk.Revit.DB.Document")
    if not isinstance(group_id, int):
        raise TypeError("group_id must be an instance of integer")

    return_value = res.Result()
    failure_messages = get_warnings_by_guid(doc, guid=guid)
    # get curve sets from failure message with matching group id
    warning_curve_sets = get_curves_from_failure_messages(
        doc=doc,
        failure_messages=failure_messages,
        group_id=group_id,
    )

    # maximum number of loops
    max_loop = len(warning_curve_sets)
    if max_loop > 0:
        return_value.append_message(
            "Found {} warnings matching group id: {}".format(max_loop, group_id)
        )

        ignore_these_curves = []
        for i in range(max_loop):
            # update the call back function with any progress made
            if callback:
                callback.update(i, max_loop)
            # start again, this time to change curve geometry
            failure_messages = get_warnings_by_guid(doc, guid=guid)
            # get next curve sets from failure message
            warning_curve_sets = get_curves_from_failure_messages(
                doc=doc,
                failure_messages=failure_messages,
                group_id=group_id,
            )
            return_value.append_message(
                "Found {} failure sets applying to group id: {}.".format(
                    len(warning_curve_sets), group_id
                )
            )
            for failure_set in warning_curve_sets:
                # check if needs to be ignored ( already tried to process before )
                curve_identifier = (failure_set[0].id, failure_set[1].id)
                # check if this set needs processing or can be ignored
                if curve_identifier not in ignore_these_curves:
                    # get geometry calcs
                    (
                        curves_to_change,
                        curves_to_delete,
                    ) = _identify_curve_in_set_to_amend_lengthening(
                        curve_set=failure_set, group_id=group_id
                    )
                    # check if None values ( cant change curves ) came back
                    if curves_to_change == None and curves_to_delete == None:
                        return_value.append_message(
                            "Neither curve {} , {} in curve set could be amended...".format(
                                failure_set[0].id, failure_set[1].id
                            )
                        )
                        # make sure to flag this set as not to be processed again
                        ignore_these_curves.append(curve_identifier)
                    else:
                        # check if any curves to change came back
                        if curves_to_change:
                            return_value.append_message(
                                "Found a curve: {} to amend".format(
                                    curves_to_change[0].id
                                )
                            )
                            # amend curves first....
                            status_update = _modify_separation_lines(
                                doc=doc,
                                curves=curves_to_change,
                                transaction_manager=transaction_manager,
                            )
                            return_value.update(status_update)
                        else:
                            return_value.append_message(
                                "Set: {} of group {} and {} of group {} contains no curve to change geometry of.".format(
                                    failure_set[0].id,
                                    failure_set[0].group_id,
                                    failure_set[1].id,
                                    failure_set[1].group_id,
                                )
                            )
                        # check if anything needs deleting
                        if curves_to_delete:
                            return_value.append_message(
                                "Found a curve: {} to delete".format(
                                    curves_to_delete[0].id
                                )
                            )
                            # delete the now overlapping curves
                            status_delete = delete_curves(
                                doc, curves_to_delete, "separation line"
                            )
                            return_value.update(status_delete)
                        else:
                            return_value.append_message(
                                "Set: {} of group {} and {} of group {} contains no curve to delete.".format(
                                    failure_set[0].id,
                                    failure_set[0].group_id,
                                    failure_set[1].id,
                                    failure_set[1].group_id,
                                )
                            )
                    # get out of loop and start again by getting the current set of warnings from the model
                    break

            # check if cancelled
            if callback:
                if callback.is_cancelled():
                    return_value.append_message("User cancelled!")
                    break
    else:
        return_value.append_message(
            "No overlap warnings with guid: {} in model.".format(guid)
        )
    return return_value


def _identify_curves_to_amend_short(curve_set, group_id):
    """
    Returns a lists containing 2 curves to change the geometry of

    :param curve_set: A set of curves that needs to be analyzed and identified for amendment
    :type curve_set: [:class: `RevitWarningOverlap`]
    :return: A tuple containing two lists - curves_to_amend and curves_to_delete
    :param group_id: group id curves need to belong to for processing (-1 not in a group)
    :type group_id: int
    :rtype: tuple
    :raises ValueError: If the length of curve_set is not equal to 2
    """

    if not isinstance(curve_set, list):
        raise TypeError("curve_set must be a list")
    if not isinstance(group_id, int):
        raise TypeError("group_id must be an instance of integer")

    curves_to_amend = []
    if len(curve_set) == 2:
        curve_to_change = calculate_shortened_curve_geometry(curve_set[0], curve_set[1])
        # check whether changes can be executed
        # need to be in main model, same design option, same phase created and demoed
        if (
            curve_to_change
            and curve_set[0].group_id == curve_set[1].group_id
            and curve_set[0].design_option_id == curve_set[1].design_option_id
            and curve_set[0].phase_created_id == curve_set[1].phase_created_id
            and curve_set[0].phase_demolished_id == curve_set[1].phase_demolished_id
        ):
            curves_to_amend.append(curve_to_change)
        else:
            curves_to_amend.append(None)
    else:
        raise ValueError("Curve set does not have a length of 2")
    return curves_to_amend


def modify_curves_by_shortening(doc, guid, transaction_manager, group_id, callback):
    """
    Modifies curves in a Revit model by lengthening them.

    :param doc: The Revit model document.
    :type doc:  (Autodesk.Revit.DB.Document)
    :param guid: The GUID used to filter the failure messages.
    :type guid: (Autodesk.Revit.DB.Guid):
    :param transaction_manager: The transaction manager used to perform the modifications.
    :type transaction_manager: (function)

    Returns:
        res.Result: A `res.Result` object that contains the status of the modifications performed.
    """

    if not isinstance(doc, Document):
        raise TypeError("doc must be an instance of Autodesk.Revit.DB.Document")
    if not isinstance(group_id, int):
        raise TypeError("group_id must be an instance of integer")

    return_value = res.Result()
    failure_messages = get_warnings_by_guid(doc, guid=guid)
    # get curve sets from failure message with matching group id
    warning_curve_sets = get_curves_from_failure_messages(
        doc=doc,
        failure_messages=failure_messages,
        group_id=group_id,
    )

    # maximum number of loops
    max_loop = len(warning_curve_sets)
    if max_loop > 0:
        return_value.append_message(
            "Found {} warnings matching group id: {}".format(max_loop, group_id)
        )
        ignore_these_curves = []
        for i in range(max_loop):
            # update the call back function with any progress made
            if callback:
                callback.update(i, max_loop)
            # start again, this time to change curve geometry
            failure_messages = get_warnings_by_guid(doc, guid=guid)
            # get next curve sets from failure message
            warning_curve_sets = get_curves_from_failure_messages(
                doc=doc,
                failure_messages=failure_messages,
                group_id=group_id,
            )
            return_value.append_message(
                "Found {} failure sets applying to group id: {}.".format(
                    len(warning_curve_sets), group_id
                )
            )
            for failure_set in warning_curve_sets:
                # check if needs to be ignored ( already tried to process before )
                curve_identifier = (failure_set[0].id, failure_set[1].id)
                # check if this set needs processing or can be ignored
                if curve_identifier not in ignore_these_curves:
                    # get geometry calcs
                    curves_to_change = _identify_curves_to_amend_short(
                        curve_set=failure_set, group_id=group_id
                    )
                    # check if None values ( cant change curves ) came back
                    if curves_to_change == None:
                        return_value.append_message(
                            "Neither curve {} , {} in curve set could be amended...".format(
                                failure_set[0].id, failure_set[1].id
                            )
                        )
                        # make sure to flag this set as not to be processed again
                        ignore_these_curves.append(curve_identifier)
                    else:
                        return_value.append_message(
                            "Found a curve: {} to amend".format(curves_to_change[0].id)
                        )
                        # amend curves first....
                        status_update = _modify_separation_lines(
                            doc=doc,
                            curves=curves_to_change,
                            transaction_manager=transaction_manager,
                        )
                        return_value.update(status_update)
                    # get out of loop and start again by getting the current set of warnings from the model
                    break

            # check if cancelled
            if callback:
                if callback.is_cancelled():
                    return_value.append_message("User cancelled!")
                    break
    else:
        return_value.append_message(
            "No overlap warnings with guid: {} in model.".format(guid)
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains utility function(s) for CAD link reports. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Revit.Links.cad_links import get_all_cad_link_instances
from duHast.Utilities import files_io as fileIO


def get_cad_link_type_data_by_name(cad_link_name, doc, revit_file_path):
    """
    Extract the file path from CAD link type.
    :param cad_link_name: The cad link name
    :type cad_link_name: str
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_file_path: The fully qualified file path to the model.
    :type revit_file_path: str
    :return: The fully qualified file path if the cad link type is a valid external reference.\
        Otherwise it will return 'unknown'.
    :rtype: str
    """

    # default values
    model_path = "unknown"
    for p in rdb.FilteredElementCollector(doc).OfClass(rdb.CADLinkType):
        if rdb.Element.Name.GetValue(p) == cad_link_name:
            try:
                ex_file_ref = None

                # trying to get the external file reference will throw an exception if not a file
                try:
                    ex_file_ref = p.GetExternalFileReference()
                except Exception:
                    pass
                
                # check if file reference
                if ex_file_ref is not None:
                    if (ex_file_ref.IsValidExternalFileReference(ex_file_ref)):
                        model_path = rdb.ModelPathUtils.ConvertModelPathToUserVisiblePath(
                            ex_file_ref.GetPath()
                        )
                        model_path = fileIO.convert_relative_path_to_full_path(
                            model_path, revit_file_path
                        )
                else:
                    # try external resource instead
                    ex_resource_references = p.GetExternalResourceReferences()
                    # check if that returned anything...
                    # the path value will be empty if the link is not loaded
                    for ref in ex_resource_references:
                        # these are c# key value pairs
                        if ref.Value.InSessionPath == "":
                            model_path = "unknown since link is unloaded"
                        else:
                            model_path = ref.Value.InSessionPath
                        break
                break
            except Exception as e:
                model_path = str(e)
    return model_path


def get_cad_report_data(doc, revit_file_path):
    """
    Gets CAD link data to be written to report file.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_file_path: The fully qualified file path to the model, which is added to data returned.
    :type revit_file_path: str
    :return: list of list of cad link properties.
    :rtype: list of list of str
    """

    data = []
    collector = get_all_cad_link_instances(doc)
    for c in collector:
        # get the workset
        ws_param = c.get_Parameter(rdb.BuiltInParameter.ELEM_PARTITION_PARAM)
        # get the design option
        do_param = c.get_Parameter(rdb.BuiltInParameter.DESIGN_OPTION_ID)
        # get the link name, link type name and shared coordinates (true or false)
        l_name_param = c.get_Parameter(rdb.BuiltInParameter.IMPORT_SYMBOL_NAME)
        # get the draw layer
        l_draw_layer_param = c.get_Parameter(rdb.BuiltInParameter.IMPORT_BACKGROUND)
        # get shared location?
        # lSharedParam = cadLink.get_Parameter(BuiltInParameter.GEO_LOCATION)
        is_view_specific = c.ViewSpecific
        owner_view_id = c.OwnerViewId
        # this returns a single string containing the cad path
        link_type_data = get_cad_link_type_data_by_name(
            l_name_param.AsString(), doc, revit_file_path
        )
        # append cad information as a list
        data.append(
            [
                revit_file_path,
                str(c.Id),
                str(l_name_param.AsString()),
                str(is_view_specific),
                str(owner_view_id),
                str(ws_param.AsValueString()),
                str(do_param.AsString()),
                str(c.Pinned),
                str(l_draw_layer_param.AsValueString()),
                link_type_data,
            ]
        )
    return data



# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import ElementId


from duHast.Revit.ColourFillSchemes.Objects.colour_fill_storage import ColourFillStorage
from duHast.Revit.ColourFillSchemes.colour_fill_scheme_entry import get_entry_value_storage_type


def get_report_data_of_colour_fill_scheme(doc, colour_fill_scheme):

    """
    Retrieves the report data from a given colour fill scheme.

    :param colour_fill_scheme: The colour fill scheme element from which to retrieve report data.
    :type colour_fill_scheme: Autodesk.Revit.DB.ColorFillScheme
    
    :return: A dictionary containing the report data.
    :rtype: dict
    """

    # get the report data
    report_data = []

    # get the area scheme id
    area_scheme_id = colour_fill_scheme.AreaSchemeId
    
    area_scheme_name = "Colour By Room Property"
    if area_scheme_id.Equals(ElementId.InvalidElementId) == False:
        # get the area scheme name
        area_scheme = doc.GetElement(area_scheme_id)
        area_scheme_name = area_scheme.Name 
    
    # get all entries in the colour fill scheme
    entries = colour_fill_scheme.GetEntries()

    # loop over the entries and get the report data
    for entry in entries:
        entry_data = ColourFillStorage()
        entry_data.fill_scheme_name = colour_fill_scheme.Name
        entry_data.area_scheme_name = area_scheme_name

        # get the value and storage type of the entry
        value_type = get_entry_value_storage_type(entry)
        entry_data.parameter_value = value_type[0]
        entry_data.storage_type = value_type[1]

        entry_data.fill_pattern_id = entry.FillPatternId.IntegerValue
        entry_data.is_in_use = entry.IsInUse
        entry_data.is_visible = entry.IsVisible
        entry_data.colour_red = entry.Color.Red
        entry_data.colour_green = entry.Color.Green
        entry_data.colour_blue = entry.Color.Blue
        report_data.append(entry_data)

        
    return report_data


# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#
import os

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_io import file_move, get_file_name_without_ext

from duHast.Revit.Family.Utility.family_copy_directive_utils import get_copy_directives

from duHast.Revit.Family.Data.Objects.family_directive_copy import FamilyDirectiveCopy
from duHast.Utilities.files_get import get_files_single_directory
from duHast.Utilities.files_io import file_exist

# post process tasks
# - move original families into back up directory
# - copy updated families into library directory


def move_files(files, destination_directory, output):
    """
    Moves files to the specified destination directory.

    :param files: List of file paths to be moved.
    :type files: list[str]
    :param destination_directory: Directory where the files will be moved.
    :type destination_directory: str
    :param output: Output function to log messages.
    :type output: function
    
    :return: True if all files were moved successfully, False otherwise.
    :rtype: bool
    """
    overall_move_flag = True
    for file in files:
        file_name =  get_file_name_without_ext(file)
        destination_file = os.path.join(destination_directory, file_name + file[-4:])  # keep the original file extension
        move_flag = file_move(file, destination_file)

        output("Moving file: {} to {} with result: {}".format(file, destination_file,  move_flag))
        overall_move_flag = overall_move_flag and move_flag
    
    return overall_move_flag


def get_part_atom_files_from_families(file_path):
    """
    Gets part atom files from the specified family file path.

    :param file_path: list of Path to the family files.
    :return: List of part atom files.
    """
    part_atom_files = []

    for family_path in file_path:
        part_atom_file_path = family_path.replace(".rfa", ".xml")
        if file_exist(part_atom_file_path):
            part_atom_files.append(part_atom_file_path)
    
    return part_atom_files


def move_original_families_to_backup_directory(family_copy_directive_directory, backupPath, output):
    """
    Moves original families from the library path to the backup path.

    :param family_copy_directive_directory: Directory containing family copy directives.
    :type family_copy_directive_directory: str
    :param backupPath: Path to the backup directory where families will be moved.
    :type backupPath: str
    :param output: Output function to log messages.
    :type output: function

    :return: Result object indicating success or failure of the operation.
    :rtype: Result
    """

    return_value = Result()

    try:
        if not os.path.exists(backupPath):
            os.makedirs(backupPath)
            return_value.append_message("Backup directory created: {}".format(backupPath))
    except Exception as e:
        return_value.update_sep(False, "Failed to create backup directory: {}".format(str(e)))
        return return_value
    
    output("Backup directory: {}".format(backupPath))

    # read copy directives
    files = get_files_single_directory(
        family_copy_directive_directory,  
        FamilyDirectiveCopy.COPY_DIRECTIVE_FILE_NAME_PREFIX, 
        "", 
        FamilyDirectiveCopy.COPY_DIRECTIVE_FILE_EXTENSION
    )
        
    if len(files) == 0:
        return_value.update_sep(False, "No copy directives found in directory: {}".format(family_copy_directive_directory))
        return return_value
    
    output("Found copy directives files: {}".format(len(files)))

    # get copy directives
    copy_directives = get_copy_directives(files)
    if copy_directives is None or len(copy_directives) == 0:
        return_value.update_sep(False, "Failed to get copy directives from files: {}".format(files))
        return return_value
    
    output("Copy directives: {}".format(len(copy_directives)))

    # build list of unique source files
    files_to_copy = []

    for copy_directive in copy_directives:
        if copy_directive.source_file_path not in files_to_copy:
            files_to_copy.append(copy_directive.source_file_path)
    
    # add part atom files if they exists
    part_atom_path = get_part_atom_files_from_families(files_to_copy)
    if part_atom_path is not None and len(part_atom_path) > 0:
        files_to_copy += part_atom_path

    # attempt to move files to backup directory
    try:

        move_flag_rfa = move_files(files_to_copy, backupPath, output)
        if not move_flag_rfa:
            return_value.update_sep(False, "Failed to move original families to backup directory: {}".format(backupPath))
        else:
            return_value.update_sep(True, "Moved original families to backup directory: {}".format(backupPath))

    except Exception as e:
        return_value.update_sep(False, "Failed to move original families to backup directory: {}".format(e))
        return return_value
    
    return return_value


def copy_new_families_to_library(output_path, library_path, output):
    """
    Copies new families from the output path to the library path.
    :param output_path: Path where new families are located.
    :type output_path: str
    :param library_path: Path to the library where families will be copied.
    :type library_path: str
    :param output: Output function to log messages.
    :type output: function

    :return: Result object indicating success or failure of the operation.
    :rtype: Result
    """

    return_value = Result()

    try:
        # get all rfa and text files
        files = get_files_single_directory(output_path, "", "",".rfa")
        files += get_files_single_directory(output_path, "", "",".txt")
        if len(files) == 0:
            return_value.update_sep(False, "No new families found in output directory: {}".format(output_path))
            return return_value

        # copy files to library path
        move_flag = move_files(files, library_path, output)
        
        if not move_flag:
            return_value.update_sep(False, "Failed to copy new families to library: {}".format(library_path))
            return return_value
        else:
            return_value.update_sep(True, "Copied new families to library: {}".format(library_path))
        return return_value
    except Exception as e:
        return_value.update_sep(False, "Failed to copy new families to library: {}".format(str(e)))
        return return_value
    

def post_process_family(output_path, library_path, backup_directory_path, output):
    """
    Post-process function to handle the final steps after family processing.

    This function copies new families to the library and moves original families to a backup directory.

    :param output_path: Path where new families are located.
    :type output_path: str
    :param library_path: Path to the library where new families will be copied to.
    :type library_path: str
    :param backup_directory_path: Path to the backup directory where original families will be moved.
    :type backup_directory_path: str

    :return: Result object indicating the success or failure of the post-processing step.
    :rtype: Result
    """

    return_value = Result()
    try:
        copy_result = copy_new_families_to_library(output_path, library_path, output)
        if copy_result.status is False:
            return_value.update_sep(
                False,
                "Failed to copy new families to library: {}".format(copy_result.message),
            )
            output("Failed to copy new families to library: {}".format(copy_result.message))
        else:
            return_value.append_message(
                "Successfully copied new families to library: {}".format(library_path)
            )
            output("Successfully copied new families to library: {}".format(library_path))
        
        backup_result = move_original_families_to_backup_directory(
            output_path, backup_directory_path, output
        )

        if backup_result.status is False:
            return_value.update_sep(
                False,
                "Failed to move original families to backup directory: {}".format(backup_result.message),
            )
            output("Failed to move original families to backup directory: {}".format(backup_result.message))
        else:
            return_value.append_message(
                "Successfully moved original families to backup directory: {}".format(backup_directory_path)
            )
            output("Successfully moved original families to backup directory: {}".format(backup_directory_path))

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to complete post process: {}".format(e),
        )
        output("Failed to complete post process: {}".format(e))
    
    output("Finished!")
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Module containing reporting functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- families

"""

# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# required for .ToList() on FilteredElementCollector
import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

# import Autodesk
from Autodesk.Revit.DB import BuiltInCategory, BuiltInParameter, Element, ElementClassFilter,FamilyInstance
from System.Collections.Generic import List

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.Objects.timer import Timer

from duHast.Revit.Common.revit_version import get_revit_version_number
from duHast.Revit.Family.Utility import loadable_family_categories as rFamUtilCats
from duHast.Revit.Family.family_utils import (
    get_family_symbols,
    get_family_instances_by_symbol_type_id,
    is_shared_from_symbol
)
from duHast.Revit.Common.parameter_get_utils import get_parameter_value_by_name
from duHast.UI.Objects.ProgressBase import ProgressBase

# default list of parameters to report on
FAMILY_PARAMETERS_TO_REPORT = [
    "Sample Parameter One",
    "Sample Parameter Two",
    "Sample Parameter Three",
]

from duHast.Revit.Family.Reporting.Objects.family_report_data import (
    FamilyReportData,
    UNKNOWN_HOST_STATUS,
)


def _get_type_properties_of_interest(
    family_symbol, parameter_names_filter, family_container
):
    """
    Get type properties matching property names in filter.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param parameter_names_filter: List of parameter nams of which to get values.
    :type parameter_names_filter: [str]
    :param family_container: The family data container.
    :type family_container: :class:`.FamilyReportData`
    """

    # get all type parameters of interest
    for parameter_name in parameter_names_filter:
        parameter_value = get_parameter_value_by_name(family_symbol, parameter_name)
        family_container.add_type_property({parameter_name: parameter_value})


def _get_instances_placed(doc, family_symbol):
    """
    Return number of instances in the model

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param family_symbol: A family type
    :type family_symbol: Autodesk.Revit.DB.FamilySymbol

    :return: Number of instances placed.
    :rtype: int
    """

    collector_instances_placed = get_family_instances_by_symbol_type_id(
        doc, family_symbol.Id
    )
    return len(collector_instances_placed.ToList())


def _get_instances_placed_revised(doc, family_symbol):
    """
    Return number of instances in the model ( fast!)

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param family_symbol: A family type
    :type family_symbol: Autodesk.Revit.DB.FamilySymbol

    :return: Number of instances placed.
    :rtype: int
    """

    filter = ElementClassFilter(FamilyInstance)
    dependent_element_ids = family_symbol.GetDependentElements(filter)

    # this is an IList, therefore I should be able to return Count
    return dependent_element_ids.Count


def _get_host_family_status(doc, family_symbol):
    """
    Returns a list of nested shared families within host family

    This will only work if there is an instance placed in the model....

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param family_symbol: A family type
    :type family_symbol: Autodesk.Revit.DB.FamilySymbol

    :return: List of family names and type names in format family name - type name
    :rtype: [str]
    """

    nested_family_names = []
    # get all instances in the model

    filter = ElementClassFilter(FamilyInstance)
    instances_element_ids = family_symbol.GetDependentElements(filter)
    #instances = get_family_instances_by_symbol_type_id(doc, family_symbol.Id)

    # get the first one
    for family_instance_id in instances_element_ids:
        family_instance = doc.GetElement(family_instance_id)
        # check if any nested shared families are in play
        try:
            sub_element_ids = family_instance.GetSubComponentIds()
            if sub_element_ids is not None:
                for sub_element_id in sub_element_ids:
                    try:
                        # get the family name and type name
                        nested_instance = doc.GetElement(sub_element_id)
                        nested_type = nested_instance.Symbol
                        nested_family = nested_type.Family
                        name = "{}-{}".format(
                                Element.Name.GetValue(nested_family),
                                Element.Name.GetValue(nested_type),
                            )
                        if name not in nested_family_names:
                            nested_family_names.append(name)
                    except Exception as e:
                        pass
        except Exception as e:
            # some family categories do not have GetSubComponentIds() available, i.e. Tags
            pass
        
        # get out
        break
        
    return nested_family_names


def report_loaded_families(
    doc, parameter_names_filter=FAMILY_PARAMETERS_TO_REPORT, progress_callback=None
):
    """
    Reports on loaded families:

    - family name
    - family type
    - family category
    - specific family type parameters and their values
    - family instances placed by by type in model


    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param output: A function piping messages to designated target.
    :type output: func(message)

    :return:
        Result class instance.

        - result.status False if an exception occurred, otherwise True.
        - result.message will contain processing messages.
        - result.result family data array

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.
        - result.result will be an empty list

    :rtype: :class:`.Result`
    """

    return_value = Result()

    # check callback class
    if progress_callback and isinstance(progress_callback, ProgressBase) == False:
        raise TypeError(
            "Output needs to be inherited from ProgressBase. Got : {} instead.".format(
                type(progress_callback)
            )
        )

    # give some initial update
    if progress_callback:
        progress_callback.update(0, 1, message="Reporting families...start")

    # build list of all categories we want families to be reloaded of
    famCats = List[BuiltInCategory](rFamUtilCats.CATEGORIES_LOADABLE_TAGS)
    famCats.AddRange(rFamUtilCats.CATEGORIES_LOADABLE_TAGS_OTHER)
    famCats.AddRange(rFamUtilCats.CATEGORIES_LOADABLE_3D)
    famCats.AddRange(rFamUtilCats.CATEGORIES_LOADABLE_3D_OTHER)

    #set up a timer
    t=Timer()
    t.start()
    # check Revit version and if 2022 and later, add new categories
    revit_version = get_revit_version_number(doc=doc)

    # log progress
    return_value.append_message("Retrieved Revit version {}. {}".format(revit_version, t.stop()))

    if revit_version >= 2022:
        famCats.AddRange(rFamUtilCats.CATEGORIES_LOADABLE_3D_REVIT_2022)
        famCats.AddRange(rFamUtilCats.CATEGORIES_LOADABLE_TAGS_REVIT_2022)

    t.start()
    # get all symbols in file
    family_symbols = get_family_symbols(doc, famCats)

    # log progress
    return_value.append_message("Got symbols in model. {}".format(t.stop()))
    
    # get families from symbols and filter out in place families
    data = []

    try:
        revit_project_file_name = doc.Title

        # get some progress data
        fam_counter = 1
        max_fam_counter = len(family_symbols.ToList())


        # loop over all family types in the model
        for family_symbol in family_symbols:
            # update progress
            if progress_callback:
                progress_callback.update(fam_counter, max_fam_counter)

            # build new data entry
            family_container = FamilyReportData()
            
            if family_symbol.Family.IsInPlace == False:
                 # get the project name
                family_container.project_name = revit_project_file_name
                # get the family name
                family_container.family_name = Element.Name.GetValue(family_symbol.Family)
                # get the type name
                family_container.family_type_name = Element.Name.GetValue(family_symbol)

                t_symbol = Timer()
                t_symbol.start()

                # get type properties
                _get_type_properties_of_interest(
                    family_symbol=family_symbol,
                    parameter_names_filter=parameter_names_filter,
                    family_container=family_container,
                )

                # log progress
                return_value.append_message("{} {} Got properties of interest: {}".format(family_container.family_name, family_container.family_type_name, t_symbol.stop()))
                t_symbol.start()

                # get number of instances placed
                family_container.family_instances_placed = _get_instances_placed_revised(
                    doc=doc, family_symbol=family_symbol
                )
                # log progress
                return_value.append_message("{} {} Got number of instances placed: {}".format(family_container.family_name, family_container.family_type_name, t_symbol.stop()))
                
                # get the family category
                family = doc.GetElement(family_symbol.Family.Id)
                family_container.family_category = family.FamilyCategory.Name

                t_symbol.start()
                # shared
                family_container.is_shared = is_shared_from_symbol(
                    family_symbol=family_symbol
                )

                # log progress
                return_value.append_message("{} {} Got is shared: {}".format(family_container.family_name, family_container.family_type_name,t_symbol.stop()))
                t_symbol.start()

                # check if this is a host family
                # this works only if a family instance is placed in the model :(
                if family_container.family_instances_placed > 0:
                    nested_shared_family_names = _get_host_family_status(
                        doc=doc,
                        family_symbol=family_symbol,
                    )
                    if len(nested_shared_family_names) > 0:
                        # record the nested family names
                            for fam_name in nested_shared_family_names:
                                family_container.add_nested_family(family_name=fam_name)
                elif (
                    family_container.family_instances_placed == 0
                ):
                    # put up note that status could not be determined...
                    # for now...
                    family_container.add_nested_family(UNKNOWN_HOST_STATUS)
                
                # log progress
                return_value.append_message("{} {} Got nested fams: {}".format(family_container.family_name, family_container.family_type_name,t_symbol.stop()))

                # add to return list
                data.append(family_container)

                # update progress:
                fam_counter = fam_counter + 1

                # check for user cancel
                if progress_callback != None:
                    if progress_callback.is_cancelled():
                        return_value.append_message("User cancelled!")
                        break

    except Exception as e:
        return_value.update_sep(
            False, "Failed to gather family data with exception: {}".format(e)
        )

    return_value.result = data
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to creating new revit revisions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# import Autodesk
import Autodesk.Revit.DB as rdb
from duHast.Utilities.Objects import result as res
from duHast.Revit.Revisions.sequence import get_revision_seq_of_name


def new_revision_action_2022(doc, revision_data):
    """
    Creates a new revision in a revit document up to version 2022.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revision_data: Named tuple containing revision data
    :type revision_data: :class:`.revisionData`

    :return:
        Result class instance.

        - Revision created status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the message revision created successfully.
        - result.result: will contain list with single entry: the new revision created

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.

    :rtype: :class:`.Result`
    """

    action_return_value = res.Result()
    try:
        new_revision = rdb.Revision.Create(doc)
        new_revision.Description = revision_data.description
        new_revision.IssuedBy = revision_data.issued_by
        new_revision.IssuedTo = revision_data.issued_to
        new_revision.NumberType = revision_data.revision_number_type
        new_revision.RevisionDate = revision_data.revision_date
        new_revision.Visibility = (
            revision_data.tag_cloud_visibility
        )  # rdb.RevisionVisibility.Hidden
        action_return_value.result.append(new_revision)
        action_return_value.update_sep(True, "Created new revision in document.")
    except Exception as e:
        action_return_value.update_sep(
            False,
            "Failed to create new revision in document with exception: " + str(e),
        )
    return action_return_value


def new_revision_action_2023(doc, revision_data):
    """
    Creates a new revision in a revit document version 2023 and onwards.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revision_data: Named tuple containing revision data
    :type revision_data: :class:`.revisionData`

    :return:
        Result class instance.

        - Revision created status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the message revision created successfully.
        - result.result: will contain list with single entry: the new revision created

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.

    :rtype: :class:`.Result`
    """

    action_return_value = res.Result()
    try:
        new_revision = rdb.Revision.Create(doc)
        new_revision.Description = revision_data.description
        new_revision.IssuedBy = revision_data.issued_by

        # set the revision sequence based on revision number type property past in
        seq = doc.GetElement(new_revision.RevisionNumberingSequenceId)
        if seq.Name == revision_data.revision_number_type:
            action_return_value.append_message(
                "Using rev sequence type assigned by default: {}".format(seq.Name)
            )
        else:
            action_return_value.append_message(
                "Need to assign none default revision sequence type: {}".format(
                    revision_data.revision_number_type
                )
            )
            rev_sequence = get_revision_seq_of_name(
                doc, revision_data.revision_number_type
            )
            if rev_sequence != None:
                new_revision.RevisionNumberingSequenceId = rev_sequence.Id
                action_return_value.append_message(
                    "Successfully assigned non default revision sequence type: {}".format(
                        rev_sequence.Name
                    )
                )
            else:
                action_return_value.append_message(
                    "Rev sequence type: {} does not exist in model. Default used instead: {} !".format(
                        revision_data.revision_number_type, seq.Name
                    )
                )

        new_revision.IssuedTo = revision_data.issued_to
        new_revision.RevisionDate = revision_data.revision_date
        new_revision.Visibility = (
            revision_data.tag_cloud_visibility
        )  # rdb.RevisionVisibility.Hidden
        action_return_value.result.append(new_revision)
        action_return_value.update_sep(True, "Created new revision in document.")
    except Exception as e:
        action_return_value.update_sep(
            False,
            "Failed to create new revision in document with exception: {}".format(e),
        )
    return action_return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Customizable element filter tests which can be used with the custom element filter actions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

These tests are used in the custom element filter actions as an argument.

test expects:

- the element of which is the test to be performed against
- the test value

"""


# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Common import worksets as rWork

# import Autodesk
from Autodesk.Revit.DB import Element


def value_in_name(value, element):
    """
    Check if provided value is in the element name.

    :param value: test value
    :type value: str
    :param element: The element
    :type element: Autodesk.Revit.DB.Element

    :return: True if part of the element name, otherwise False
    :rtype: bool
    """

    return value in Element.Name.GetValue(element)


def value_equals_workset_name(value, element):
    """
    Check if provided value is equal to the elements workset name.

    :param value: test value
    :type value: str
    :param element: The element
    :type element: Autodesk.Revit.DB.Element

    :return: True if elements workset name matches, otherwise False
    :rtype: bool
    """

    return rWork.is_element_on_workset_by_name(element, value)


def value_in_family_name(value, element):
    """
    Check if provided value is part of the elements Family name.

    :param value: test value
    :type value: str
    :param element: The element
    :type element: Autodesk.Revit.DB.Element

    :return: True if elements family name partly matches, otherwise False
    :rtype: bool
    """

    return value in Element.Name.GetValue(element.Symbol.Family)


def value_in_element_type_family_name(value, element):
    """
    Check if provided value is part of the element type Family name.

    :param value: test value
    :type value: str
    :param element: The element
    :type element: Autodesk.Revit.DB.Element

    :return: True if elements family name partly matches, otherwise False
    :rtype: bool
    """

    return value in element.Symbol.FamilyName


def value_is_family_name(value, element):
    """
    Check if provided value is equal to the elements Family name.

    :param value: test value
    :type value: str
    :param element: The element
    :type element: Autodesk.Revit.DB.Element

    :return: True if elements family name matches, otherwise False
    :rtype: bool
    """

    return value == Element.Name.GetValue(element.Symbol.Family)


def value_is_element_type_family_name(value, element):
    """
    Check if provided value is equal to the element type Family name.

    :param value: test value
    :type value: str
    :param element: The element
    :type element: Autodesk.Revit.DB.Element

    :return: True if elements family name matches, otherwise False
    :rtype: bool
    """

    return value == element.Symbol.FamilyName



# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#
import os

from duHast.Utilities.files_io import rename_file
from duHast.Utilities.files_get import get_files_single_directory
from duHast.Utilities.Objects import result as res
from duHast.Utilities.directory_io import directory_exists
from duHast.UI.Objects.ProgressBase import ProgressBase
from duHast.Revit.Exports.Utility.convert_pdf_dwg_settings import convert_settings_json_string_to_settings_objects
from duHast.Revit.Exports.Utility.export_utility import replace_illegal_characters_from_dwg_file_name


from System.Collections.Generic import List

from Autodesk.Revit.DB import ACADVersion, BaseExportOptions,Document, DWGExportOptions,ElementId,ViewSheet

from duHast.Revit.Exports.Utility.export_utility import get_sheet_parameter_data


def set_dwg_export_option():

    # Create a new instance of DWGExportOptions
    dwg_export_options = DWGExportOptions()

    dwg_export_options.FileVersion = ACADVersion.R2010
    dwg_export_options.HideReferencePlane = True
    dwg_export_options.HideScopeBox = True
    dwg_export_options.HideUnreferenceViewTags	= True
    dwg_export_options.MergedViews = True

    return dwg_export_options


def get_all_dwg_export_options(doc):
    """
    Retrieves all DWG export options.

    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of DWG export options.
    :rtype: list
    """

    # Get the export options for the specified setup name
    dwg_export_options = DWGExportOptions.GetPredefinedOptions(doc)

    # Get the export option names
    setup_names = BaseExportOptions.GetPredefinedSetupNames(doc)

    return dwg_export_options

def get_dwg_export_option_by_name(doc, dwg_export_option_name):
    """
    Retrieves the DWG export option by name.

    :param dwg_export_option_name: The name of the DWG export option.
    :type dwg_export_option_name: str
    :return: The DWG export option.
    :rtype: Autodesk.Revit.DB.DWGExportOptions
    """
    
    setup_names = BaseExportOptions.GetPredefinedSetupNames(doc)

    for setup_name in setup_names:
        if setup_name == dwg_export_option_name:
            # Get the export options for the specified setup name
            dwg_export_options = DWGExportOptions.GetPredefinedOptions(doc, setup_name)
            return dwg_export_options

    return None


def export_sheet_to_dwg (doc, view_sheet, sheet_name_string, output_directory, dwg_export_option_name=None):
    """
    Exports a Revit sheet to PDF using the provided naming rule and output directory.

    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param view_sheet: The sheet to export.
    :type view_sheet: Autodesk.Revit.DB.ViewSheet
    :param sheet_name_string: The string to build the sheet name.
    :type sheet_name_string: str
    :param output_directory: The directory to save the PDF file.
    :type output_directory: str
    """

    return_value = res.Result()
    try:

        # Check if we got a document
        if not isinstance(doc, Document):
            raise TypeError("doc must be an instance of Autodesk.Revit.DB.Document")
    
        # Check if we got a list of sheets
        if not isinstance(view_sheet,  ViewSheet):
            raise TypeError("view_sheet must be an Autodesk.Revit.DB.ViewSheet instances")
        
        # Check if the output directory exists
        if not directory_exists(output_directory):
            return_value.update_sep(False, "Output directory does not exist.")
            return return_value

        # Check if the sheet name string is a string or None
        if not isinstance(sheet_name_string, str) and sheet_name_string is not None:
            raise TypeError("sheet_name_string must be a string or None")
        
        # check if the dwg export option name is a string or None
        if dwg_export_option_name is not None and not isinstance(dwg_export_option_name, str):
            raise TypeError("dwg_export_option_name must be a string or None")


        dwg_export_option = None

        if dwg_export_option_name:
            dwg_export_option = get_dwg_export_option_by_name(doc, dwg_export_option_name)
            if dwg_export_option is None:
                return_value.update_sep(False, "DWG export option {} not found.".format(dwg_export_option_name))
                return return_value
        else:
            # Set the naming rule if provided
            dwg_export_option = set_dwg_export_option()
    
        # convert to .net list
        sheets = List[ElementId]()
        sheets.Add(view_sheet.Id)

        # Export the sheet to PDF
        export_result = doc.Export(
            output_directory,
            "",
            sheets,
            dwg_export_option
        )

        # reporting..
        sheet_identifier = "{} {}".format(view_sheet.SheetNumber, view_sheet.Name)

        # Check if the export was successful
        if export_result:
            return_value.update_sep(True, "Exported sheet {} to DWG successfully.".format(sheet_identifier))
        else:
            return_value.update_sep(False, "Export of sheet {} failed.".format(sheet_identifier))

        # If the export was successful, rename the file
        if export_result:
            # need to rename the cad file after export...
            # Get the export file name
            exported_file_name = view_sheet.SheetNumber + " - " + view_sheet.Name

            # remove the invalid characters from the file name
            # this is the name Revit uses to export the file (Revit replaces characters that are 'not allowed' in file names with dashes)
            exported_file_name = replace_illegal_characters_from_dwg_file_name(exported_file_name)
            
            # find the file
            files_match = get_files_single_directory(
                folder_path= output_directory, 
                file_prefix="", 
                file_suffix=exported_file_name, 
                file_extension=".dwg"
            )

            # check if the file exists
            if len(files_match) > 1:
                # more than one file found
                return_value.update_sep(False, "More than one file found matching {} . Please check the output directory.".format(exported_file_name))
                return return_value
            elif len(files_match) == 0:
                # no file found
                return_value.update_sep(False, "No file found matching {} . Please check the output directory.".format(exported_file_name))
                return return_value

            # Get the sheet parameter data
            sheet_parameter_data = get_sheet_parameter_data(view_sheet)

            # convert name rules string to settings objects
            dwg_name_rules = convert_settings_json_string_to_settings_objects(sheet_name_string)
            
            # set up new sheet name
            sheet_name_new = []

            # set up a rule counter to keep track of whether a rule is the last of the rules
            rule_counter = 0

            # loop over the naming rules
            for rule in dwg_name_rules:

                if rule.propertyName not in sheet_parameter_data:
                    # if the rule is not a parameter name, skip it
                    continue

                # check if the rule has a prefix
                if rule.prefix is not None:
                    sheet_name_new.append(rule.prefix)

                # add the parameter value to the sheet name
                sheet_name_new.append(sheet_parameter_data[rule.propertyName])

                # check if the rule has a suffix
                if rule.suffix is not None:
                    sheet_name_new.append(rule.suffix)

                # only add the separator if it is not None and not the last one
                if rule.separator is not None and rule_counter < len(dwg_name_rules):
                    sheet_name_new.append(rule.separator)

                # increase the rule counter
                rule_counter += 1

            # build the sheet name from the chunks list
            sheet_name_new_joined = "".join(sheet_name_new)

            # remove the invalid characters from the file name
            # but keep the full stop if any to match the doc number / name in Revit
            sheet_name_new_joined = replace_illegal_characters_from_dwg_file_name(sheet_name_new_joined, replace_full_stop=False)

            # built the new file name
            new_file_name = os.path.join(output_directory , sheet_name_new_joined + ".dwg")

            # Rename the file
            rename_result = rename_file(files_match[0], new_file_name)

            if(rename_result):
                return_value.append_message("...File renamed successfully to match naming rule: {}".format(sheet_name_new_joined))
            else:
                return_value.update_sep(False, "...File rename to {} failed.".format(sheet_name_new_joined))

        else:
            # if the export failed, we don't need to rename the file
            return_value.append_message("...Export failed, no need to rename the file.")

        return return_value
    
    except Exception as e:
        # handle the exception
        return_value.update_sep(False, "Error exporting sheets to dwg: {}".format(str(e)))
        return return_value
    

def export_sheets_to_dwg(doc, sheets, sheet_name_string, output_directory, dwg_export_option_name=None, callback=None):
    """
    Exports a list of Revit sheets to DWG using the provided naming rule and output directory.

    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param sheets: The list of sheets to export.
    :type sheets: list
    :param sheet_name_string: The string to build the sheet name.
    :type sheet_name_string: str
    :param output_directory: The directory to save the DWG file.
    :type output_directory: str
    :param dwg_export_option_name: The name of the DWG export option.
    :type dwg_export_option_name: str
    :param callback: The callback to update the progress.
    :type callback: ProgressBase

    :return: The result of the export operation.
    :rtype: Result
    """

    return_value = res.Result()

    try:

        # Check if we got a document
        if not isinstance(doc, Document):
            raise TypeError("doc must be an instance of Autodesk.Revit.DB.Document")
    
        # Check if we got a list of sheets
        if not isinstance(sheets, list) or not all(isinstance(sheet, ViewSheet) for sheet in sheets):
            raise TypeError("sheets must be a list of Autodesk.Revit.DB.ViewSheet instances")
        
        # Check if the output directory exists
        if not directory_exists(output_directory):
            return_value.update_sep(False, "Output directory does not exist.")
            return return_value

        # Check if the sheet name string is a string or None
        if not isinstance(sheet_name_string, str) and sheet_name_string is not None:
            raise TypeError("sheet_name_string must be a string or None")
        
        # check if the dwg export option name is a string or None
        if dwg_export_option_name is not None and not isinstance(dwg_export_option_name, str):
            raise TypeError("dwg_export_option_name must be a string or None")
    
        # check if the callback is of progressBase
        if callback is not None and not isinstance(callback, ProgressBase):
            raise TypeError("callback must be an instance of ProgressBase or None")
        
        # Check if the callback is None, if so export all in one go
        if callback is None:

            # loop over sheet at the time and export to pdf
            for i, sheet in enumerate(sheets):

                # Export the sheet to PDF
                export_result = export_sheet_to_dwg(doc, sheet, sheet_name_string, output_directory, dwg_export_option_name)
                return_value.update(export_result)
           
        else:
            # export one sheet at the time and update the progress
            total_sheets = len(sheets)
            callback.update(0,  total_sheets)

            # loop over sheet at the time and export to pdf
            for i, sheet in enumerate(sheets):

                # update the progress
                callback.update(i, total_sheets)

                # Export the sheet to PDF
                export_result = export_sheet_to_dwg(doc, sheet, sheet_name_string, output_directory, dwg_export_option_name)
                return_value.update(export_result)

                # check if user cancel the export
                if callback.is_cancelled():
                    return_value.append_message("User cancelled!")
                    break
    
    except Exception as e:
        # handle the exception
        return_value.update_sep(False, "Error exporting sheets to DWG: {}".format(str(e)))
        return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A class used to return status, messages and objects back to a caller.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The `Result` class is a subclass of the `base.Base` class and is used to store and manipulate the results of a process. It has fields for `message`, `status`, and `result`, and provides methods to update these fields.

Example Usage:
    result = Result()  # Create a new instance of the Result class
    result.append_message("An error occurred")  # Append a new message to the existing message
    result.update_status(False)  # Update the status to False
    result.update_sep(True, "Process completed successfully")  # Update the status and append a new message
    print(result)  # Print the result object

Methods:
    __init__(self):
        Initializes the Result object with default values for `message`, `status`, and `result`

    __repr__(self):
        Returns a string representation of the Result object, including the formatted message, status, and result

    append_message(self, message):
        Appends a new line and new message string to the existing message

    update(self, otherResult):
        Updates the Result object using another Result object, updating the message, status, and result

    update_sep(self, status, message):
        Updates the status and message properties of the Result object

    update_status(self, status):
        Updates the status property of the Result object

Fields:
    message: A string representing the message of the result
    status: A boolean representing the status of the result
    result: A list to store the result items
   
"""


#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects import base


class Result(base.Base):
    def __init__(self):
        """
        Class constructor.

        - message default value is -
        - status default value is True
        - result default value is []

        """

        super(Result, self).__init__()

        self.status = True
        self.result = []
        self._messages = [] # performance improvement to store messages as a list

    def __repr__(self):

        # Use str.format() to add indentation and line breaks
        formatted_string = "\n".join(["...{}".format(line) for line in self._messages])

        return "message: \n{} \nstatus: [{}] \nresult: {}".format(
            formatted_string, self.status, self.result
        )

    def append_message(self, message):
        """
        Appends a  new message string to the existing messages list.

        :param message: The new message to be appended.
        :type message: str
        """

        if (isinstance(message,str) == False):
            raise TypeError("message must be an instance of string")
        
       
        self._messages.append(message)


    @property
    def message(self):

        if len(self._messages) == 0:
            return "-"
        else:
            return "\n".join(self._messages)
    
    @property
    def message_as_list(self):
        return self._messages

    def update(self, otherResult):
        """
        Will use the past in result instance to update the instance.

        - .status is using a logical AND
        - .message is using append (unless other message is default '-')
        - .result is looping over past in result and adding it one by one to this .result list (ignores None)

        :param otherResult: Another result class instance.
        :type otherResult: SampleBatchProcessorCode.Result
        """

        if not isinstance(otherResult, Result):
            raise TypeError("otherResult must be an instance of Result")

        try:
            # check if default message string, if so do not update
            if len(otherResult.message_as_list)>0 :
                self._messages = self._messages + otherResult.message_as_list
            self.status = self.status & otherResult.status
            # check if result property that was passed in has values
            if any(otherResult.result):
                for item in otherResult.result:
                    self.result.append(item)
        except Exception as e:
            print(str(e))
            pass

    def update_sep(self, status, message):
        """
        Updates the .status and .message property only.

        - .status is using a logical AND
        - .message is using append (unless other message is default '-')

        :param status: The status to be added.
        :type status: bool
        :param message: The message to be appended.
        :type message: str
        """

        if not isinstance(status, bool):
            raise TypeError("status must be an instance of boolean")

        try:
            self.append_message(message)
            self.status = self.status & status
        except Exception as e:
            print(str(e))
            pass

    def update_status(self, status):
        """
        Update .status only.

        - .status is using a logical AND

        :param status: The status to be added.
        :type status: bool
        """

        try:
            self.status = self.status & status
        except Exception as e:
            print(str(e))
            pass



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit API utility functions for the spatial properties of room elements.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#!/usr/bin/python
# -*- coding: utf-8 -*-
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

from duHast.Revit.Rooms.Objects.RoomBaseObject import RoomBaseObj
from duHast.Revit.Rooms.Objects.RoomSpatialObject import RoomSpatialObj
from Autodesk.Revit.DB import (
    Line,
    SpatialElementBoundaryLocation,
    XYZ,
)

from duHast.Utilities.unit_conversion import convert_imperial_feet_to_metric_mm
from duHast.Revit.Common.Geometry.curve import are_lines_parallel, are_lines_perpendicular

class RoomSpatialForView(RoomBaseObj):
    def __init__(
        self, rvt_doc, room, boundary_location=SpatialElementBoundaryLocation.Finish
    ):
        # initialize the base class
        super(RoomSpatialForView, self).__init__(rvt_doc, room, boundary_location=boundary_location)

        # Use the helper method to calculate spatial data
        (self.segments, self.room_walls, self.wall_segs, 
         self.bbox, self.bbox_centre) = self._calculate_spatial_data(rvt_doc, room, boundary_location)


    @staticmethod
    def _calculate_spatial_data(rvt_doc, room, boundary_location):
        """
        Helper method to compute spatial data for a room.

        This is done because I cant directly inherit from RoomSpatialObj due to exception:
        IronPython.Runtime.Exceptions.TypeErrorException: metaclass conflict: the metaclass of a derived class must be a (non-strict) subclass of the metaclasses of all its bases
        which is caused by .net metaclass conflict.

        """

        # Create a dummy room spatial object to get some spatial data
        dummy_room = RoomSpatialObj(rvt_doc, room, boundary_location=boundary_location)

        return dummy_room.segments, dummy_room.room_walls, dummy_room.wall_segs, dummy_room.bbox, dummy_room.bbox_centre
    

    def is_room_rectalinear(self):
        """
        Check if the room is rectalinear. ( all room bounding segements are either parallel or perpendicular to each other )

        :return: True if the room is rectalinear, False otherwise.
        :rtype: bool
        """

        #  IList<IList<Autodesk.Revit.DB.BoundarySegment>> segments = room.GetBoundarySegments(new SpatialElementBoundaryOptions());
        # get all room bounding segments ( this is a list of list of BoundarySegment objects )
        room_bounding_segemnts = self.segments

        # proceed only if room is bound:
        if room_bounding_segemnts is None:
            return False
        
        
        # check if all room bounding segments are either parallel or perpendicular to each other
        # if so, return True else return False

        # loop over outer loop segments only ( representing the outer room boundary and not any inner islands ) and check if parallel or perpendicular to the previous segment
        room_bounding_segemnts_outer_loop = room_bounding_segemnts[0]
        #print ("Room bounding segments loops: ", len(room_bounding_segemnts))
        #print("Room bounding segments outer loop: ", len(room_bounding_segemnts_outer_loop))
        for i in range(1, len(room_bounding_segemnts_outer_loop)):
            # get the current segment
            current_segment_curve = room_bounding_segemnts_outer_loop[i].GetCurve()
            #print("...Current segment curve: ", current_segment_curve)
            #print("...Current segment type: ", type(room_bounding_segemnts_outer_loop[i]))
            # get the previous segment
            previous_segment_curve = room_bounding_segemnts_outer_loop[i-1].GetCurve()
            #print("...Previous segment curve: ",previous_segment_curve)
            #print("...Previous segment type: ", type(room_bounding_segemnts_outer_loop[i-1]))

            # make sure both curves are lines ( no arcs or other curves )
            if (isinstance(current_segment_curve, Line) and isinstance(previous_segment_curve, Line)):
                # check if the current segment is parallel or perpendicular to the previous segment
                if(not(are_lines_parallel(current_segment_curve, previous_segment_curve) or are_lines_perpendicular(current_segment_curve, previous_segment_curve))):
                    # if not, the room is not rectalinear
                    #print("......Current segment is parallel {} \n{}{}" .format(are_lines_parallel(current_segment_curve, previous_segment_curve), current_segment_curve.Direction.Normalize(), previous_segment_curve.Direction.Normalize()))
                    #print("......Current segment is perpendicular {} \n{}" .format(are_lines_perpendicular(current_segment_curve, previous_segment_curve), current_segment_curve.Direction.Normalize().DotProduct(previous_segment_curve.Direction.Normalize())))
                    
                    print("...length of current segment: ", convert_imperial_feet_to_metric_mm(current_segment_curve.Length))
                    
                    return False
            else:
                # only lines are supported
                #print("Only lines are supported for checking if room is rectalinear")
                return False
            
        # if all segments are either parallel or perpendicular to each other, the room is rectalinear
        return True
    

    def is_room_aligned_to_its_bounding_box(self):
        """
        Check if the room is aligned to the bounding box. ( all room bounding segements are either parallel or perpendicular to the bounding box )

        :return: True if the room is aligned to the bounding box, False otherwise.
        :rtype: bool
        """

        # check if room is bound:
        if self.bbox == None:
            return False
        
        # check if room is rectalinear
        if not(self.is_room_rectalinear()):
            # check if any random room bounding segments is parallel or perpendicular to the bounding box
            return False
        

        # check if any arbritary room bounding segment is parallel or perpendicular to any bounding box edge
        if(self.segments is not None and len(self.segments) > 0):
            # get the first room bounding segment
            room_bounding_segemnts_outer_loop = self.segments[0]
            first_segment_curve = room_bounding_segemnts_outer_loop[0].GetCurve()
            # check if the first segment is parallel or perpendicular to any bounding box edge
            # get a bounding box edge from min XYZ and max XYZ
            # by using Line.CreateBound
            bbox_edge = Line.CreateBound(self.bbox.Min, XYZ(self.bbox.Max.X, self.bbox.Min.Y, self.bbox.Min.Z))
            # check if the first segment is parallel or perpendicular to the bounding box edge
            if(are_lines_parallel(first_segment_curve, bbox_edge) or are_lines_perpendicular(first_segment_curve, bbox_edge)):
                return True
            else:
                return False
        else:
            return False
